//////////////////////////////////////////////////////////////////////////////
// Sketch: usb_ghost
// แสดงตัวอย่างการจำลองบอร์ดไมโครคอนโทรลเลอร์ให้เป็นอุปกรณ์คีย์บอร์ด เมาส์ และเกมแพด
// ในเวลาเดียวกัน
//
// == การคอมไพล์ ==
// ต้องการไลบรารี pt และ usbdrv (มีมาให้กับแพคเกจ cpeku-practicum เรียบร้อยแล้ว)
//
// == การใช้งาน ==
// ทำแสงบริเวณตัวรับแสงให้มืดเพื่อสั่งให้บอร์ดเริ่มส่งคีย์และขยับเมาส์
//
// == หมายเหตุ ==
// อุปกรณ์ USB ที่จำลองขึ้นกำหนดให้มีหมายเลข VID:PID เป็น 16c0:0482 ซึ่งได้รับการจดทะเบียน
// ให้ใช้งานเป็น Keyboard/Mouse/Joystick combo โดย VOTI
//////////////////////////////////////////////////////////////////////////////
#include <pt.h>
#include <usbdrv.h>
#include "keycodes.h"

#define REPORT_ID_KEYBOARD 1
#define REPORT_ID_MOUSE    2
#define REPORT_ID_GAMEPAD  3

// มาโครสำหรับจำลองการหน่วงเวลาใน protothread
#define PT_DELAY(pt,ms,tsVar) \
  tsVar = millis(); \
  PT_WAIT_UNTIL(pt, millis()-tsVar >= (ms));

/////////////////////////////////////////////////////////////////////
// USB report descriptor, สร้างขึ้นจาก HID Descriptor Tool ซึ่ง
// ดาวน์โหลดได้จาก
//    http://www.usb.org/developers/hidpage/dt2_4.zip
//
// หรือใช้ HIDEdit ซึ่งให้บริการฟรีผ่านเว็บที่ http://hidedit.org/
//
// *** ไม่แนะนำให้สร้างเองด้วยมือ ***
/////////////////////////////////////////////////////////////////////
PROGMEM const char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = 
{
  ///////////////////
  // Keyboard
  ///////////////////
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x09, 0x06,                    // USAGE (Keyboard)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x85, 0x01,                    //   REPORT_ID (1)
  0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
  0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
  0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //   REPORT_SIZE (1)
  0x95, 0x08,                    //   REPORT_COUNT (8)
  0x81, 0x02,                    //   INPUT (Data,Var,Abs)
  0x95, 0x01,                    //   REPORT_COUNT (1)
  0x75, 0x08,                    //   REPORT_SIZE (8)
  0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
  0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
  0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
  0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
  0xc0,                          // END_COLLECTION

  ///////////////////
  // 3-Button Mouse
  ///////////////////
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x0b, 0x02, 0x00, 0x01, 0x00,  // USAGE (Generic Desktop:Mouse)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x0b, 0x01, 0x00, 0x01, 0x00,  //   USAGE (Generic Desktop:Pointer)
  0xa1, 0x00,                    //   COLLECTION (Physical)
  0x85, 0x02,                    //     REPORT_ID (2)
  0x05, 0x09,                    //     USAGE_PAGE (Button)
  0x1b, 0x01, 0x00, 0x09, 0x00,  //     USAGE_MINIMUM (Button:Button 1)
  0x2b, 0x03, 0x00, 0x09, 0x00,  //     USAGE_MAXIMUM (Button:Button 3)
  0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //     REPORT_SIZE (1)
  0x95, 0x03,                    //     REPORT_COUNT (3)
  0x81, 0x02,                    //     INPUT (Data,Var,Abs)
  0x75, 0x05,                    //     REPORT_SIZE (5)
  0x95, 0x01,                    //     REPORT_COUNT (1)
  0x81, 0x03,                    //     INPUT (Cnst,Var,Abs)
  0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
  0x0b, 0x30, 0x00, 0x01, 0x00,  //     USAGE (Generic Desktop:X)
  0x0b, 0x31, 0x00, 0x01, 0x00,  //     USAGE (Generic Desktop:Y)
  0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
  0x25, 0x7f,                    //     LOGICAL_MAXIMUM (127)
  0x75, 0x08,                    //     REPORT_SIZE (8)
  0x95, 0x02,                    //     REPORT_COUNT (2)
  0x81, 0x06,                    //     INPUT (Data,Var,Rel)
  0xc0,                          //     END_COLLECTION
  0xc0,                          // END_COLLECTION

  ////////////////////////////////////////////////////
  // Gamepad with one button and one analog stick
  ////////////////////////////////////////////////////
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x09, 0x04,                    // USAGE (Joystick)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x05, 0x01,                    //   USAGE_PAGE (Generic Desktop)
  0x09, 0x01,                    //   USAGE (Pointer)
  0xa1, 0x00,                    //   COLLECTION (Physical)
  0x85, 0x03,                    //     REPORT_ID (3)
  0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
  0x09, 0x32,                    //     USAGE (Z)
  0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
  0x26, 0xff, 0x03,              //     LOGICAL_MAXIMUM (1023)
  0x75, 0x0a,                    //     REPORT_SIZE (10)
  0x95, 0x01,                    //     REPORT_COUNT (1)
  0x81, 0x02,                    //     INPUT (Data,Var,Abs)
  0x05, 0x09,                    //     USAGE_PAGE (Button)
  0x09, 0x01,                    //     USAGE (Button 1)
  0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //     REPORT_SIZE (1)
  0x95, 0x01,                    //     REPORT_COUNT (1)
  0x81, 0x02,                    //     INPUT (Data,Var,Abs)
  0x75, 0x01,                    //     REPORT_SIZE (1)
  0x95, 0x05,                    //     REPORT_COUNT (5)
  0x81, 0x03,                    //     INPUT (Cnst,Var,Abs)
  0xc0,                          //   END_COLLECTION
  0xc0                           // END_COLLECTION
};

struct ReportKeyboard
{
  /* +----\------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |Byte \ Bit |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
   * +------\----+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  0        |               Report ID = 1                   |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  1        |           Modifiers (shift,ctrl,etc)          |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  2        |                 Key Code                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   */
  uint8_t  report_id;
  uint8_t  modifiers;
  uint8_t  key_code;
};

struct ReportMouse
{
  /* +----\------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |Byte \ Bit |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
   * +------\----+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  0        |               Report ID = 2                   |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  1        |              Buttons' statuses                |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  2        |                  Delta X                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  3        |                  Delta Y                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   */
  uint8_t  report_id;
  uint8_t  buttons;
  int8_t   dx;
  int8_t   dy;
};

struct ReportGamepad
{
  /*
   * +----\------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |Byte \ Bit |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
   * +------\----+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  0        |               Report ID = 3                   |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  1        |                 Light (L)                     |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  3        |            UNUSED           | Btn | Light (H) |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   */
  uint8_t  report_id;
  uint16_t light:10;  // ค่าแสง 10 บิต (0-1023)
  uint8_t  button:1;  // ปุ่ม 1 บิต
  uint8_t  unused:5;  // ไม่ใช้ แต่ต้องเติมให้เต็มไบต์
};


ReportKeyboard reportKeyboard;
ReportMouse reportMouse;
ReportGamepad reportGamepad;

uint16_t light;  // Updated by Light-Task; to be shared among threads

// Protothread states
struct pt main_pt;
struct pt keyboard_pt;
struct pt key1_pt;
struct pt key2_pt;
struct pt key3_pt;
struct pt key4_pt;
struct pt key5_pt;
struct pt key6_pt;
struct pt key7_pt;
struct pt key8_pt;
struct pt key9_pt;
struct pt key10_pt;
struct pt key11_pt;
struct pt key12_pt;

////////////////////////////////////////////////////////////////
// Automatically called by usbpoll() when host makes a request
////////////////////////////////////////////////////////////////
usbMsgLen_t usbFunctionSetup(uchar data[8])
{
  usbRequest_t *rq = (usbRequest_t*)data;

  return 0;  /* Return nothing to host for now */
}

//////////////////////////////////////////////////////////////
void sendKey(uint8_t keycode, uint8_t modifiers)
{
  reportKeyboard.key_code = keycode;
  reportKeyboard.modifiers = modifiers;
  usbSetInterrupt((uchar*)&reportKeyboard, sizeof(reportKeyboard));
}

//////////////////////////////////////////////////////////////
void sendMouse(int8_t dx, int8_t dy, uint8_t buttons)
{
  reportMouse.dx = dx;
  reportMouse.dy = dy;
  reportMouse.buttons = buttons;
  usbSetInterrupt((uchar*)&reportMouse, sizeof(reportMouse));
}

//////////////////////////////////////////////////////////////
PT_THREAD(key1_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PC0)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_L,0);
    if(digitalRead(PIN_PC0)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}
//////////////////////////////////////////////////////////////
PT_THREAD(key2_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PC1)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_K,0);
    if(digitalRead(PIN_PC1)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key3_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PC2)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_J,0);
    if(digitalRead(PIN_PC2)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key4_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PC3)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_I,0);
    if(digitalRead(PIN_PC3)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key5_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PC4)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_H,0);
    if(digitalRead(PIN_PC4)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key6_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PB0)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_D,0);
    if(digitalRead(PIN_PB0)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key7_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PB1)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_S,0);
    if(digitalRead(PIN_PB1)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key8_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PB2)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_A,0);
    if(digitalRead(PIN_PB2)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key9_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PB3)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_W,0);
    if(digitalRead(PIN_PB3)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
}//////////////////////////////////////////////////////////////
PT_THREAD(key10_task(struct pt *pt))
{
  static uint32_t ts = 0;
  PT_BEGIN(pt);
  
  while(digitalRead(PIN_PB4)==0){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_F,0);
    if(digitalRead(PIN_PB4)==1){
    PT_DELAY(pt,5,ts);
    sendKey(KEY_NONE,0);
    }  
  }
  PT_DELAY(pt,5,ts);
  PT_END(pt);
  
}
//////////////////////////////////////////////////////////////
void setup()
{
  // Initialize USB subsystem
  usbInit();
  usbDeviceDisconnect();
  delay(300);
  usbDeviceConnect();
  
  // Initialize hardware pins

  pinMode(PIN_PC3, INPUT_PULLUP);
  pinMode(PIN_PC4, INPUT_PULLUP);
  pinMode(PIN_PC5, INPUT_PULLUP);
  pinMode(PIN_PC0, INPUT_PULLUP);
  pinMode(PIN_PC1, INPUT_PULLUP);
  pinMode(PIN_PC2, INPUT_PULLUP);
  pinMode(PIN_PB3, INPUT_PULLUP);
  pinMode(PIN_PB4, INPUT_PULLUP);
  pinMode(PIN_PB5, INPUT_PULLUP);
  pinMode(PIN_PB0, INPUT_PULLUP);
  pinMode(PIN_PB1, INPUT_PULLUP);
  pinMode(PIN_PB2, INPUT_PULLUP);
  pinMode(PIN_PD3, OUTPUT);

  // Initialize USB reports
  reportKeyboard.report_id = REPORT_ID_KEYBOARD;
  reportKeyboard.modifiers = 0;
  reportKeyboard.key_code = KEY_NONE;

  

  // Initialize tasks
  PT_INIT(&key1_pt);
  PT_INIT(&key2_pt);
  PT_INIT(&key3_pt);
  PT_INIT(&key4_pt);
  PT_INIT(&key5_pt);
  PT_INIT(&key6_pt);
  PT_INIT(&key7_pt);
  PT_INIT(&key8_pt);
  PT_INIT(&key9_pt);
  PT_INIT(&key10_pt);
}

//////////////////////////////////////////////////////////////
void loop()
{
  usbPoll();
  key1_task(&key1_pt);
  key1_task(&key2_pt);
  key1_task(&key3_pt);
  key1_task(&key4_pt);
  key1_task(&key5_pt);
  key1_task(&key6_pt);
  key1_task(&key7_pt);
  key1_task(&key8_pt);
  key1_task(&key9_pt);
  key1_task(&key10_pt);
}
